module Complement where

import Data.List
import Control.Monad
import AST
import Data.Maybe (fromJust)

import Debug.Trace

import AndOrSearch

data Subst a = Subst [AVariable a] [TTree (AVariable a) a]
instance Show a => Show (Subst a) where
    showList ls s = concatMap (\s -> show s++"\n") ls ++ s
    show (Subst vs ts) = show vs ++ " --> " ++ show ts 

type EqCon a = [(TTree (AVariable a) a, TTree (AVariable a) a)]

orderEq (t1,t1') (t2,t2') =
    case orderT t1 t2 of
      LT -> LT
      GT -> GT
      EQ -> case orderT t1' t2' of
              LT -> LT
              GT -> GT
              EQ -> EQ
    where cmpT (t:ts) (t':ts') | orderT t t' == LT = LT
                               | orderT t t' == GT = GT
                               | otherwise         = cmpT ts ts'
          cmpT [] [] = EQ
          orderT (C c ts) (C c' ts') | c < c'    = LT
                                     | c > c'    = GT
                                     | otherwise = cmpT ts ts'
          orderT (V v) (V v') = compare v v'
          orderT (V v) _ = LT
          orderT _ (V v) = GT



detectOverlap eqvs substs traversed cnt limit =  
    let (b,_) = detectOverlapStep eqvs substs [] limit
    in b

-- False --> Rules have no overlaps
-- True  --> Rules may have an overlap
detectOverlapStep eqvs substs pfacts limit = -- trace (show substs) $
    if lightIndepCheck eqvs then
        (False,pfacts)
    else
        case search (And $ nub $ concatMap (\(e,e')->reduceTE e e') eqvs) gen compareEq test check include pfacts of
          (False3,   pfacts) -> (False, pfacts)
          (True3,    pfacts) -> (True , pfacts)
          (Unknown3, pfacts) -> trace (show "= Overlap Unknown ===\n"++show eqvs) (True , pfacts)
    where
      (gen, test, check, compareEq, include) = generateSearchFuncs substs limit
      lightIndepCheck eqvs = any (\(e,e') -> diffHead e e') eqvs
      diffHead (C c _) (C c' _)    | c /= c' = True
      diffHead (C c ts) (C c' ts') | c == c' = or (zipWith diffHead ts ts') 
      diffHead _ _ = False


--detectOverlapStep eqvs substs cnt limit = 
--    let (gen, test, check, compareEq) = generateSearchFuncs rules limit
--    in

include n [] = False
include n (n':traversed) =  
    if same n n' then
        True
    else
        include n traversed
    where
      same (And eqvs) (And eqvs') = same' eqvs eqvs'
      same (Or  eqvs) (Or  eqvs') = same' eqvs eqvs'
      same _ _ = False
      same' eqvs eqvs' = p eqvs eqvs' && isBj (bindL eqvs eqvs') && isBj (bindR eqvs eqvs') 
      p ((a,a'):eqvs) ((b,b'):eqvs') = eq a b && eq a' b' && p eqvs eqvs'
      p [] [] = True
      p _ _ = False
      eq (F f ts) (F f' ts') | f == f' = and (zipWith eq ts ts')
      eq (C c ts) (C c' ts') | c == c' = and (zipWith eq ts ts')
      eq (V (VFromW f _ i)) (V (VFromW f' _ i')) = f == f' && i == i'
      eq (V VAny) (V VAny) = True
      eq _ _ = False
      isBj bs = isInj bs && (isInj (map (\(i,j)->(j,i)) bs))
      isInj bs = all (\e->case e of {[v]->True;_->False}) $ groupBy (\x y->fst x==fst y) $ sort bs
      bindL eqvs eqvs' = nub $ concat $ zipWith b (map fst eqvs) (map fst eqvs')
      bindR eqvs eqvs' = nub $ concat $ zipWith b (map snd eqvs) (map snd eqvs')       
      b (C c ts) (C c' ts') | c == c' = concat $ zipWith b ts ts'
      b (V (VFromW f j i)) (V (VFromW f' j' i')) | f == f' && i == i' = [(j,j')]
      b (V VAny) (V VAny) = []
      b _ _ = []
              

generateSearchFuncs substs limit = 
    (gen, test, check, compareNode, include)
    where
      sizeE t = dp 0 t
          where dp i (C _ ts) = maximum (map (dp (i+1)) ts)
                dp i (V v)    = i
                dp i _        = 0
      maximum        = foldr max 0
      maxSizeE       = maximum $ map (\(Subst _ rs)->maximum $ map sizeE rs) substs
      maxSizeT       = maximum $ map (\(Subst _ rs)->length rs) substs
      sizeEq eqvs    = maximum $ map (\(e,e')->sizeE e `max` sizeE e') eqvs
      isSafe eqvs    = length eqvs <= maxSizeT && sizeEq eqvs <= maxSizeE
      compareNode (And eqvs) (And eqvs') = compareEq eqvs eqvs' 
      compareNode (Or eqvs) (Or eqvs')   = compareEq eqvs eqvs'
      compareEq eqvs eqvs' 
          | isSafe eqvs && isSafe eqvs' = compare (sizeEq eqvs) (sizeEq eqvs')
          | isSafe eqvs && not (isSafe eqvs') = LT
          | not (isSafe eqvs) && isSafe eqvs' = GT
          | otherwise                         = compare (sizeEq eqvs) (sizeEq eqvs')
      test (And []) = True
      test (Or [])  = False
      test _        = False
      check (And eqvs) dep = isSafe eqvs || dep < limit
      check (Or eqvs)  dep = isSafe eqvs || dep < limit     
      gen (And eqvs) = -- (\s -> trace (show eqvs ++ " --split to -> \n" ++ foldr (\a r->show a++"\n"++r) "\n" (sortBy compareNode s)) s) $ 
          map Or $ map (sortBy orderEq) $ splitIndependentEq $ eqvs
      gen (Or eqvs) = -- (\s -> trace (show eqvs ++ " ==> \n" ++ foldr (\a r->show a++"\n"++r) "\n" (sortBy compareNode s)) s) $
                       map And $ map (sortBy orderEq.nub) $ map unifier eqvss
          where eqvssL   = filter unifiable [ map (\(t,t') -> (sub t s j (next s+maxNumEq+1), t')) eqvs | s <- substsL, j <- candsL eqvs ]
                eqvssR   = filter unifiable [ map (\(t,t') -> (t, sub t' s j (next s+maxNumEq+1))) eqvs | s <- substsR, j <- candsR eqvs ]
                eqvss = eqvssL ++ eqvssR 
                cand (F _ ts) = concatMap cand ts
                cand (C _ ts) = concatMap cand ts
                cand (V (VFromW f rid ii)) = [rid]
                cand _ = []
                candsL eqvs = nub $ concatMap cand $ map fst eqvs
                candsR eqvs = nub $ concatMap cand $ map snd eqvs
                substsL = filter (cont (map fst eqvs)) substs
                substsR = filter (cont (map snd eqvs)) substs
                maxNumEq = maximum $ map (\(e,e') -> maxVarI e `max` maxVarI e') eqvs
                next (Subst lv rv) = (maximum $ map maxVarI rv)
                maxVarI (F f ts) = maximum (map maxVarI ts) 
                maxVarI (C c ts) = maximum (map maxVarI ts) 
                maxVarI (V (VFromW f rid ii)) = rid
                maxVarI (V v)    = 0
                maximum [] = 0
                maximum (a:x) = max a (maximum x)
                unifiable eqvs = all p eqvs
                    where p (C c ts, C c' ts') | c == c' = all p $ zip ts ts'
                          p (V v, t)                     = True
                          p (t, V v)                     = True
                          p _                            = False
                unifier eqvs = concatMap (\(t,t')->reduceTE t t') eqvs
                
{-
 Analalysis for Complementation
-}
reportAnalysis rules limit = 
    showMatrices 
    where substs = makeSubsts rules
          matrices = owMatrix rules substs limit
          fsymbs   = nub $ [ f | ARule rid f ps es _ <- rules ]
--          showInjectives = "Injectivity:\n"++
--                           s [ (f,detectInjective f rules substs limit) | f <- fsymbs, f /= idSymb ]++"\n"
--              where s = foldr (\(f,t) r->show f ++ ": " ++ show t ++ "\n"++r) []
          showMatrices = foldr (\(f,a) r->showMatrix f a++"\n"++r) [] matrices 
          showMatrix f [] = "Overlap for \""++show f++"\" is trivial.\n"
          showMatrix f (mat) = 
              strf ++ 
              replicate maxW ' ' ++ "|" ++ foldr (\a r->fillSp (show a)++" "++r) [] idxs ++ "\n"++
              replicate maxW '-' ++ "+" ++ replicate ((maxW+1)*length idxs) '-' ++ "\n"++
              foldr (\i r->
                        fillSp (show i) ++ "|" ++
                        foldr (\j r->
                                fillSp (find i j)++" "++r) ("\n"++r) idxs) [] idxs
              where
                strf = "Overlaps for \""++show f++"\":\n"
                idxs = sort $ nub $ [ k | (f,i,j,t) <- mat, k <- [i,j] ]
                find i j = case [ t | (f,i',j',t) <- mat, (i==i'&&j==j') || (i==j'&&j==i') ] of
                             [True]  -> "O"
                             [False] -> " "
                             _       -> "O"
                maxI = foldr max 0 [ max i j | (f,i,j,t) <- mat ]
                minI = foldr min maxI [ min i j | (f,i,j,t) <- mat ]
                maxW = length (show maxI)
                fillSp s = s ++ replicate (maxW-(length s)) ' '

    
owMatrix rules substs limit = map (\f -> (f,overlapMatrix f rules substs limit)) fsymbs
    where  fsymbs = nub $ [ f | ARule rid f ps es _ <- rules, f /= idSymb ]

overlapMatrix f rules substs limit = matrix
    where frules = filter (\(ARule _ f' _ _ _) -> f==f') rules          
          wids wdecls = concat [ vis vs' | WRule wid vs f vs' <- wdecls ]
              where vis (VFromW _ id _:vs) = id:vis vs
                    vis (v:vs) = vis vs
                    vis [] = []
          (pfacts,matrix) =
              foldr (\(f,rid,rid',es,es') (pfacts,rs) ->
                         let (hasOw, pfacts') = detectOverlapStep (zip es es') substs pfacts limit 
                         in ( pfacts', (f,rid,rid',hasOw):rs )) ([],[])
                        [ (f,rid,rid',es,es')
                              | ARule rid  f  ps  es  _ <- frules,
                                ARule rid' f' ps' es' _ <- frules, rid < rid']
--          matrix = [ (f,rid,rid', detectOverlap (zip es es') substs [] (max_I+1) limit)
--                         | ARule rid  f  ps  es  _ <- frules,
--                           ARule rid' f' ps' es' _ <- frules, rid < rid']


-- For Normalized Rule
detectInjective f rules substs limit = 
    fix [f] [idSymb]
    where fix [] proced = True
          fix (SNrm f []:req) proced = False
          fix (f:req) proced | f `elem` proced = fix req proced 
                             | otherwise = 
              case detectInjectiveStep f rules substs limit of 
                (False,_)    -> False
                (True, req') -> fix (req'++req) (f:proced)


detectInjectiveStep f rules substs limit = 
    (null res && all used frules, fr )
        where frules = rulesOf f rules
--              max_I  = foldr max 0 [ foldr max 0 [ i | WRule i _ _ _ <- wdecls] | ARule rid f ps es wdecls <- frules ]
              res = filter (\(f,i,j,b) -> b) $ overlapMatrix f rules substs limit 
              used (ARule _ _ _ es wdecls) = null (wvars \\  evars) 
                  where evars = nub $ concatMap variables es
                        wvars = nub $ concatMap (\(WRule _ vs _ _)->vs) wdecls
              fr = nub $ concatMap p frules
                   where p (ARule _ _ _ _ wdecls) = map (\(WRule _ _ f _)->f) wdecls

-- detectInjective f rules substs limit = detectInjectiveI f frets rules substs limit
--     where frets = [0..head [ length es | ARule i f' ps es ws <- rules, f==f']-1]
 
-- detectInjectiveI f idxs rules substs limit = fix [(f,idxs)] [(idSymb,[0])]
--     where fix [] proced = True
--           fix ((f,idxs):req) proced = 
--               case detectInjectiveStep f rules substs limit idxs proced of
--                 (False,_) -> False
--                 (True,[]) -> fix (req) ((f,idxs):proced)
--                 (True,req') -> fix (req++req') ((f,idxs):proced)
   

-- detectInjectiveStep f rules substs limit idxs proced 
--     | (f,idxs) `elem` proced = (True, [])
--     | otherwise = (all not res, fr)
--     where frules = filter (\(ARule _ f' _ _ _) -> f==f') rules
--           max_I  = foldr max 0 [ foldr max 0 [ i | WRule i _ _ _ <- wdecls] | ARule rid f ps es wdecls <- frules ]
--           res = [ let z  = zip es es' 
--                       eq = map (\i -> z !! i) idxs
--                   in detectOverlap eq substs [] (max_I+1) limit 
--                                              | (ARule rid  _ _ es  _) <- frules,
--                                                (ARule rid' _ _ es' _) <- frules, rid < rid' ]
--           fr  = nub $ concatMap p frules
--               where p (ARule _ _ ps es wdecls) = require
--                         where swdecls = sortBy (\(WRule i _ _ _) (WRule j _ _ _) -> compare i j) wdecls
--                               wids    = [ i | WRule i _ _ _ <- swdecls ]
--                               avars = nub $ concat [ activeVars (es !! i) | i <- idxs ]
--                               jidxss = map (\wid ->sort$nub$concatMap (\v->
--                                                                        case v of VFromW _ i j | i == wid -> [j]
--                                                                                  _ -> []) avars) wids
--                               require = nub $ zip wsymbols jidxss 
--                               wsymbols = [ f | WRule i _ f _ <- swdecls ]
                                                      


makeSubsts :: [ARule (AVariable a) a] -> [Subst a]
makeSubsts xs = Subst [] []:map mkSubst xs
    where mkSubst :: ARule (AVariable a) a -> Subst a
          mkSubst (ARule rid fname pts exp _) = 
              Subst (map (\i->VFromF fname i) [0..(length exp-1)]) exp

-- detectOverlap eqvs substs traversed cnt limit = fst$detectOverlapStep eqvs substs traversed [] cnt limit
-- detectOverlapStep eqvs substs traversed pfacts cnt limit = 
--     if lightIndepCheck eqvs then
--         (False,pfacts)
--     else
--         detectOverlapS (concatMap (\(e,e')->reduceTE e e') eqvs) substs traversed pfacts cnt limit 
--     where
--       lightIndepCheck eqvs = any (\(e,e') -> diffHead e e') eqvs
--       diffHead (C c _) (C c' _)    | c /= c' = True
--       diffHead (C c ts) (C c' ts') | c == c' = or (zipWith diffHead ts ts') 
--       diffHead _ _ = False
        
-- detectOverlapS eqvs substs traversed pfacts cnt limit 
--     | length traversed > limit = (True, pfacts)
--     | check eqvs pfacts        = (True, pfacts)
--     | check eqvs traversed     = (False,pfacts)
--     | True = -- (\s -> trace (show eqvs++" > "++(show s)) s) $
--         let rets = allUnifiableSubsts eqvs substs cnt 
--         in chk rets pfacts 
--           where chk [] pfact = (False, pfact)
--                 chk ((eq,cnt'):eqvss) pfact =
--                     let (hasOw, pfact') = chkIn (splitIndependentEq eq) pfact cnt'
--                     in if hasOw then
--                            (hasOw,pfact')
--                        else
--                            chk eqvss pfact'
--                 chkIn [] pfact cnt' = (True, eqvs:pfact)
--                 chkIn (eqvs':eqvss) pfact cnt' =
--                     let (hasOw, pfact') = detectOverlapS eqvs' substs (eqvs:traversed) pfact cnt' limit 
--                     in if hasOw then
--                            chkIn eqvss pfact' cnt'
--                        else
--                            (hasOw, pfact') 
-- {-       ( or [ all (\eqv->detectOverlap eqv substs (eqvs:traversed) cnt' limit) (splitIndependentEq eq) 
--                  | (eq,cnt')  <- allUnifiableSubsts eqvs substs cnt ] , pfacts ) -}



-- splitIndependentEq :: Eq a => EqCon a -> [EqCon a]
splitIndependentEq eqvs = f eqvs 
    where f [] = []
          f ((t,t'):eqvs) = let (t1,t2) = fix (wvars t, wvars t') [(t,t')] eqvs
                            in t1:f t2
          fL v ts = partition (\(t,t')->include v t) ts
          fR v ts = partition (\(t,t')->include v t') ts
          step (vs,vs') ts = 
              let (t1,t2) = foldr (\v (t1,t2) -> let (t1',t2') = fL v t2 in (t1'++t1, t2')) ([],ts) vs
                  (s1,s2) = foldr (\v (t1,t2) -> let (t1',t2') = fR v t2 in (t1'++t1, t2')) (t1,t2) vs'
              in ( s1, s2 )
          fix (vs,vs') ts1 ts2 = 
              let (t1,t2) = step (vs,vs') ts2
              in if null t1 then
                     (ts1, t2)
                 else
                     let t1' = t1++ts1
                     in fix (nub$wvarsL t1'++vs, nub$wvarsR t1'++vs') t1' t2
          wvarsL ts = concat $ map (wvars . fst) ts
          wvarsR ts = concat $ map (wvars . snd) ts
          wvars (F _ ts) = concatMap wvars ts
          wvars (C _ ts) = concatMap wvars ts
          wvars (V (VFromW f rid ii)) = [VFromW f rid ii]
          wvars (V v) = []
          include v (F _ ts) = any (include v) ts
          include v (C _ ts) = any (include v) ts
          include (VFromW f rid ii) (V (VFromW f' rid' _)) | f == f' && rid == rid' = True
          include _ _ = False

sub t (Subst lv rv) j i = sub' t
    where rv' = map replaceVars rv
          replaceVars (F f ts) = F f (map replaceVars ts)
          replaceVars (C c ts) = C c (map replaceVars ts)
          replaceVars (V (VFromW f rid ii)) = V (VFromW f (rid+i) ii)
          replaceVars (V v) = V v
          sub' (F f ts) = F f (map sub' ts)
          sub' (C c ts) = C c (map sub' ts)
          sub' (v@(V (VFromW f rid ii))) | rid == j = 
              case findIndex (chk f ii) lv of 
                Just k  -> rv' !! k
                Nothing -> v
          sub' (V v) = V v
          chk f ii (VFromF f' ii') | f == f' && ii == ii' = True
                                   | otherwise            = False
                

-- sub :: Eq a => TTree (AVariable a) a -> Subst a -> Int -> TTree (AVariable a) a
-- sub t (Subst lv rv) i = sub' t
--     where rv' = map replaceVars rv
--           replaceVars (F f ts) = F f (map replaceVars ts)
--           replaceVars (C c ts) = C c (map replaceVars ts)
--           replaceVars (V (VFromW f rid ii)) = V (VFromW f (rid+i) ii)
--           replaceVars (V v) = V v
--           sub' (F f ts) = F f (map sub' ts)
--           sub' (C c ts) = C c (map sub' ts)
--           sub' (V (VFromW f rid ii)) = 
--               case findIndex (chk f ii) lv of
--                 Just j  -> rv' !! j
--                 Nothing -> V (VFromW f rid ii)
--           sub' (V v) = V v
--           chk f ii (VFromF f' ii') | f == f' && ii == ii' = True
--                                    | otherwise            = False

cont ts (Subst lv rv) = any haveS ts && any topRed ts
    where haveS (F _ ts) = any haveS ts
          haveS (C _ ts) = any haveS ts
          haveS (V (VFromW f rid ii)) = any (chk f ii) lv  
          haveS (V v) = False
          topRed (V (VFromW f rid ii)) = any (chk f ii) lv
          topRed _ = False
          chk f ii (VFromF f' ii') | f == f' && ii == ii' = True
                                   | otherwise            = False


reduceTE (F f ts) (F f' ts') | f == f' =
    concat$ zipWith reduceTE ts ts'
reduceTE (C c ts) (C c' ts') | c == c' = 
    concat$ zipWith reduceTE ts ts'
reduceTE (V (VFromW f _ _)) (V (VFromW f' _ _)) | (f == idSymb || f' == idSymb) = [] 
reduceTE t t' = [(t,t')]              
          

-- -- Given, equivalence constraint [(t1,t1'), ...., (s1,s1')],
-- --- Find substitions theta and theta' s.t. 
-- ---   there exists mgu sigma, forall (ti,ti'), (ti theta sigma = ti' theta' sigma)
-- allUnifiableSubsts eqvs substs i = (\s -> trace (show eqvs++" > "++foldr (\a r->show a++"\n"++r) [] s) s) $ 
--     reduceResult $ 
--      filter (\(e,j) -> all (\(t,t') -> unifiable t t') e) $ 
--           [ (map (\(a,b) -> let r  = sub a s  i
--                                 r' = sub b s' i 
--                             in (r,r')) eqvs, i+next s+next s'+1) | s  <- substsL, 
--                                                                    s' <- substsR  ]  
--     where substsL = Subst [] []:filter (cont (map fst eqvs)) substs
--           substsR = Subst [] []:filter (cont (map snd eqvs)) substs
--           next (Subst lv rv) = maximum $ map maxVarI rv
--           maxVarI (F f ts) = maximum (map maxVarI ts) 
--           maxVarI (C c ts) = maximum (map maxVarI ts) 
--           maxVarI (V (VFromW f rid ii)) = rid
--           maxVarI (V v)    = 0
--           maximum [] = 0
--           maximum (a:x) = max a (maximum x)
--           reduceResult xs = nubBy (\(e,j) (e',j') -> e == e') $ 
--                               map (\(e,j) -> (concat$map (\(x,y) -> reduceTE x y) e,j)) $
--                                filter (\(e,j) -> reducable e) xs 
--           reducable xs = trace (show xs++ ":" ++ show (red xs)) $ null xs || allVars xs || red xs
--               where red ((t,t'):xs) = isR t t' || red xs
--                     red []          = False
--                     isR (F f ts) (F f' ts') | f == f' = True
--                     isR (C c ts) (C c' cs') | c == c' = True
--                     isR _ _ = False
--                     allVars ((V v,V v'):xs) = allVars xs
--                     allVars [] = True
--                     allVars _  = False
--           unif (F f ts) (F f' ts') | f == f' = 
--                                             do rs <- (zipWithM unif ts ts')
--                                                return $ concat rs
--           unif (C c ts) (C c' ts') | c == c' = 
--                                             do rs <- (zipWithM unif ts ts')
--                                                return $ concat rs
--           unif (V VAny) _ = return $ []
--           unif _ (V VAny) = return $ [] 
--           unif (V v) t = return $ [(v,t)]
--           unif t (V v) = return $ [(v,t)]
--           unif _ _ = Nothing
--           unifiable t t' = 
--               case unif t t' of
--                 Just bind ->
--                     checkBind bind
--                 Nothing ->
--                     False
--           checkBind [] = True
--           checkBind (b:bs) = 
--               case mapM (replace b) bs of
--                 Just bs' -> checkBind bs'
--                 Nothing  -> False
--           replace (v,t) (v',t') =
--               if inc v' t then
--                   Nothing
--               else
--                   return $ (v',rep (v,t) t')
--           inc v (F _ ts) = any (inc v) ts
--           inc v (C _ ts) = any (inc v) ts
--           inc v (V v')   = v == v'
--           rep (v,t) (F f ts) = F f (map (rep (v,t)) ts)
--           rep (v,t) (C c ts) = C c (map (rep (v,t)) ts)
--           rep (v,t) (V v')   = if v == v' then t else V v'
                                   

-- check _ [] = False
-- check eqvs (eqvs':traversed) = 
--     if same eqvs eqvs' then
--         True
--     else
--         check eqvs traversed
--     where same eqvs eqvs'  = same' (sortBy orderEq eqvs) (sortBy orderEq eqvs')
--           same' ((a1,a2):x) ((b1,b2):y) = eq a1 b1 && eq a2 b2 && same' x y
--           same' [] [] = True
--           same' _ _   = False
--           eq (F f ts) (F f' ts') | f == f' = and (zipWith eq ts ts')
--           eq (C c ts) (C c' ts') | c == c' = and (zipWith eq ts ts')
--           eq (V (VFromW f _ i)) (V (VFromW f' _ i')) = f == f' && i == i'
--           eq (V VAny) (V VAny) = True
--           eq _ _ = False






{-
   Complementation
-}
complementation f rules limit = trace (show $ prettyVariables tnf nrules)
    nrules'
    where frets = [0..head [ length es | ARule i f' ps es ws <- rules, f==f']-1]
          nf = SNrm f frets
          tnf = STpl (SNrm f frets)
          rules' = normalizeProgram f$introduceIdentity$removeUnusedWhere$ rules
          brules = introduceIdentity$removeUnusedWhere$basicComplementation rules' limit
          dependsMap = makeDependMap brules
          nrules  = introduceIdentity$removeUnusedWhere$removeConstructor tnf dependsMap brules limit 
          nrules' = introduceIdentity$removeUnusedWhere$renameConstructor tnf dependsMap nrules limit


checkIndependenceAll fs rules substs limit = 
    not $ checkOwAll fs []
    where 
      checkOwAll [] pfacts = False
      checkOwAll (f:fs) pfacts =
          let (hasOw, pfacts') = checkOw f rules pfacts 
          in if hasOw then True
             else checkOwAll fs pfacts'
      checkOw f rules pfacts = 
          foldr (\(es,es') (bool,pfacts)->
                     if bool then (bool,pfacts)
                     else detectOverlapStep (zip es es') substs pfacts limit) (False,pfacts) $                          
                    [ (es,es') | ARule rid  _ _ es  _ <- frules,
                                 ARule rid' _ _ es' _ <- frules, rid < rid' ]
              where frules = rulesOf f rules
                    max_I  = foldr max 0 [ foldr max 0 (wids wdecls) | ARule rid f ps es wdecls <- frules ]
                    wids wdecls = concat [ vis vs' | WRule wid vs f vs' <- wdecls ]
                        where vis (VFromW _ id _:vs) = id:vis vs
                              vis (v:vs) = vis vs
                              vis [] = []

                    
makeDependMap rules = fixImpr fmaps
    where fsymbs = nub $ [ f | (ARule _ f _ _ _) <- rules ]
          fmaps  = map (\f->(f,sort$nub $ f:[ g | (ARule _ g _ _ ws) <- rules, f `inRHS` ws ])) fsymbs
          impr fm = map (\(f,fs)->( f, sort $ nub $ [ h | f' <- fs, (g,gs)<-fm, g==f',h<-gs] )) fm
          fixImpr fm = 
              let fm' = impr fm 
              in if fm' == fm then fm
                 else fixImpr fm'
          inRHS f ws = any (\(WRule _ _ g _) -> g==f) ws


renameConstructor f dependsMap rules limit = fix [f] rules []
    where fix [] rls proc = rls 
          fix (f:req) rls proc 
              | f `elem` proc = fix req rls proc 
              | otherwise = 
                  let ( rls', req') = renameConstructorStep f (fromJust$lookup f dependsMap) rls limit 
                  in fix (req++req') rls' (f:proc) 

renameConstructorStep f dependFuncs rules limit = 
    ( merge [] frules [] 0++nfrules, require )
    where (frules, nfrules) = partition (\(ARule _ f' _ _ _) -> f==f') rules
          merge prs [] rcpairs cnt = prs
          merge prs (r:frules) rcpairs cnt = 
              let (r',rcpair',cnt') = oneStep r (frules++prs++nfrules) rcpairs cnt
              in merge (r':prs) frules rcpair' cnt'
          makeMatrix rules = concatMap (\(f,i,j,t)->if t then [(i,j),(j,i)] else [] ) $
                                overlapMatrix f rules (makeSubsts rules) limit 
          matrix = makeMatrix frules
          hasCompatibleConstructor matrix (ARule i (STpl f) _ es _) (ARule j _ _ es' _) = 
               if ( not $ (i,j) `elem` matrix ) then  
                   all (\(e,e')->case (e,e') of
                                   (C c cs, C c' cs') | c /= c' -> length cs == length cs'
                                   _ -> False) $ drop (length es `div` 2) (zip es es')
               else
                   False
          hasCompatibleConstructor _ _ _ = False
          changeConstructor c (ARule i (STpl f) ps es ws) = (ARule i (STpl f) ps es' ws)
              where l = length es `div` 2 
                    es' = take l es++
                          map (\e-> case e of 
                                      C c' cs' -> C (SSys$c++"_"++(show$length cs'))  cs'
                                      _ -> e) (drop l es)                                                 
          isConstructorForm (ARule i (STpl f) ps es ws) =  
               all (\e ->  case e of 
                             C c cs -> True
                             _ -> False) $ drop (length es `div` 2) es
          isConstructorForm _ = False
          oneStep r rest rcpairs cnt = 
              if isConstructorForm r then
                  check [ (r',c') | (r',c') <- rcpairs, hasCompatibleConstructor matrix r' r ] r rest cnt
              else
                  (r,rcpairs,cnt)
              where
                check [] r rest cnt = 
                    let c' = "B"++show cnt
                        r' = changeConstructor c' r
                        rules' = r':rest
                    in if checkIndependenceAll dependFuncs rules' (makeSubsts rules') limit then
                           (r',(r',c'):rcpairs,cnt)
                       else 
                           check [] r rest (cnt+1)
                check ((r',c'):rcp) r rest cnt =  
                    let r'     = changeConstructor c' r
                        rules' = r':rest 
                    in if checkIndependenceAll dependFuncs rules' (makeSubsts rules') limit then
                           (r',rcpairs,cnt)
                       else 
                           check rcp r rest cnt
          require = nub $ concatMap (\(ARule i f' ps es ws)->map (\(WRule i vs f vs')->f) ws) frules 


-- renameConstructorStep f dependFuncs rules limit = -- trace ("Checking:"++show f++ " with "++show dependFuncs) $ 
--     (merge [] frules [] 0++nfrules, require)
--     where frules   = rulesOf f rules
--           nfrules  = rulesNotOf f rules
--           modifierpos es = length es `div` 2
--           makeMatrix rules = concatMap (\(f,i,j,t)->if t then [(i,j),(j,i)] else [] ) $
--                                 overlapMatrix f rules (makeSubsts rules) limit 
--           matrix = makeMatrix frules
--           hasCompatibleConstructor matrix (ARule i (STpl f ids) _ es _) (ARule j _ _ es' _) = 
--                if ( not $ (i,j) `elem` matrix ) then  
--                    all (\(e,e')->case (e,e') of
--                                    (C c cs, C c' cs') | c /= c' -> length cs == length cs'
--                                    _ -> False) $
--                       map (\(i,e,e')->(e,e')) $filter (\(i,e,e')->(i-modifierpos es) `elem` ids)$zip3 [0..] es es'  
--                else
--                    False
--           hasCompatibleConstructor _ _ _ = False
--           changeConstructor c (ARule i (STpl f ids) ps es ws) = (ARule i (STpl f ids) ps es' ws)
--               where es' = zipWith (\i e-> if (i-modifierpos es) `elem` ids then
--                                               case e of 
--                                                 C c' cs' -> C (SSys$c++"_"++(show$length cs'))  cs'
--                                                 _ -> e
--                                           else 
--                                               e) [0..] es                                                 
--           isConstructorForm (ARule i (STpl f ids) ps es ws) =  
--                all (\(i,e) -> (not$(i-modifierpos es) `elem` ids) || case e of 
--                                                                 C c cs -> True
--                                                                 _ -> False) $ zip [0..] es
--           isConstructorForm _ = False
-- --           groupRules matrix [] = []
-- --           groupRules matrix (r:rs) = 
-- --               let (n,p) = partition (hasCompatibleConstructor matrix r) rs
-- --               in (r:n):groupRules matrix p
--           merge prs [] rcpairs cnt = prs
--           merge prs (r:frules) rcpairs cnt = 
--               let (r',rcpair',cnt') = oneStep r (frules++prs++nfrules) rcpairs cnt
--               in merge (r':prs) frules rcpair' cnt'
--           oneStep r rest rcpairs cnt = 
--               if isConstructorForm r then 
--                   check [ (r',c') | (r',c') <- rcpairs, hasCompatibleConstructor matrix r' r ] r rest cnt 
--               else
--                   (r,rcpairs,cnt)
--               where
--                 check [] r rest cnt = 
--                     let c' = "B"++show cnt
--                         r' = changeConstructor c' r
--                         rules' = r':rest
--                     in if checkIndependenceAll dependFuncs rules' (makeSubsts rules') limit then
--                            (r',(r',c'):rcpairs,cnt)
--                        else 
--                            check [] r rest (cnt+1)
--                 check ((r',c'):rcp) r rest cnt =  
--                     let r'     = changeConstructor c' r
--                         rules' = r':rest 
--                     in if checkIndependenceAll dependFuncs rules' (makeSubsts rules') limit then
--                            (r',rcpairs,cnt)
--                        else 
--                            check rcp r rest cnt
-- --           oneStep [] prs rest cnt = (prs,[])
-- --           oneStep (r:grp) prs rest cnt = 
-- --               if isConstructorForm r then 
-- --                   let r'     = changeConstructor ("B"++show cnt) r
-- --                       rules'  = r':(grp++rest)
-- --                       substs' = makeSubsts rules'
-- --                   in if checkIndependenceAll dependFuncs rules' substs' limit then
-- --                          oneStep grp (r':prs) rest cnt
-- --                      else
-- --                          (prs, r:grp)
-- --               else
-- --                   (r:prs, grp)
--           require = nub $ concatMap (\(ARule i f' ps es ws)->map (\(WRule i vs f vs')->f) ws) frules 


removeConstructor f dependsMap rules limit = fix [f] rules []
    where fix [] rls proc = rls 
          fix (f:req) rls proc 
              | f `elem` proc = fix req rls proc 
              | otherwise =
                  let ( rls', req') = removeConstructorStep f (fromJust$lookup f dependsMap) rls limit 
                  in fix (req++req') rls' (f:proc) 

removeConstructorStep f dependFuncs rules limit = 
    (check (mfrules++nfrules) sfrules rsfrules, require)
    where frules  = rulesOf f rules        -- Rules of f
          nfrules = rulesNotOf f rules     -- Rules that is not rule of f
          singletonRule (ARule i f ps es ws) = 
              case f of
                (STpl _) -> 
                    let es' = drop (length es `div` 2) es
                    in case es' of 
                         (C c [t]):_ -> True
                         _   -> False
                _ -> False
          (sfrules, mfrules) = partition singletonRule frules
          rsfrules = map (\(ARule i f ps es ws)->
                              let l   = length es `div` 2 
                                  esO  = take l es
                                  esC' = map (\e -> case e of {C c [t] -> t ; _ -> e}) (drop l es)
                              in ARule i f ps (esO++esC') ws) sfrules
          check rls [] [] = rls
          check rls (s:srules) (r:rrules) = trace (show ()++ " " ++ show (checkIndependenceAll dependFuncs rules' substs' limit )) $ 
              if checkIndependenceAll dependFuncs rules' substs' limit then
                  check (r:rls) srules rrules 
              else
                  check (s:rls) srules rrules  
              where rules'  = r:(srules++rls)
                    substs' = makeSubsts rules'
          require = concatMap (\(ARule i f' ps es ws)->map (\(WRule i vs f vs')->f) ws) frules 
                    
                     
-- removeConstructorStep f dependFuncs rules limit =  
--     (check (mfrules++nfrules) sfrules rsfrules False, require)
--       where frules  = rulesOf f rules
--             nfrules = rulesNotOf f rules
--             modifierpos es = length es `div` 2
--             singletonRule (ARule i f ps es ws) =
--                 case f of                   
--                   (STpl s (x:xs)) -> 
--                       case es !! (x + modifierpos es) of
--                         C c [t] -> True
--                         _       -> False
--                   _ -> False
--             sfrules = filter singletonRule frules
--             mfrules = filter (not.singletonRule) frules
--             rsfrules = map (\(ARule i (STpl s is) ps es ws)->
--                                     ARule i (STpl s is) ps
--                                               (zipWith (\i e -> if (i-modifierpos es) `elem` is then
--                                                                     case e of 
--                                                                       C c [t] -> t
--                                                                       _ -> e
--                                                                 else
--                                                                     e ) [0..] es) ws) sfrules
--             check rls [] [] b = rls
--             check rls (s:srules) (r:rrules) b = 
--                 let rules'  = rls++(r:srules)
--                     substs' = makeSubsts rules'
--                 in if checkIndependenceAll dependFuncs rules' substs' limit then
--                        check (r:rls) srules rrules True
--                    else 
--                        check (s:rls) srules rrules b
--             require = concatMap (\(ARule i f' ps es ws)->map (\(WRule i vs f vs')->f) ws) sfrules 



-- Assume that rules are already normalized
basicComplementation rules limit = 
    let ( rules', rID ) = 
            foldr (\r (rs,cnt)->(basicComplementationStep r rules substs limit cnt:rs,cnt+1))
                      ([],maximumRID rules+1) rules
    in rules'
    where 
      substs = makeSubsts rules


basicComplementationStep (ARule rid f ps es wdecls) rules substs limit cnt 
    | f == idSymb = ARule rid f ps es wdecls
    | otherwise = 
        ARule cnt (STpl f) ps (map chVar es++map conv es) wdecls'
    where
      swdecls = sortBy (\(WRule i _ _ _) (WRule j _ _ _) -> compare i j) wdecls
      wids     = [ i | WRule i _ _ _ <- swdecls ]
      warities = [ length ps | WRule i ps _ _ <- swdecls ]
      wsymbols = [ f | WRule i _ f _ <- swdecls ]
      injs  = map (\g-> detectInjective g rules substs limit) wsymbols
      chVar (C c ts)              = C c (map chVar ts)
      chVar (F f ts)              = F f (map chVar ts)
      chVar (V (VFromW f i j)) | f == idSymb = V (VFromW f i j) -- Special Treatment for ID
                               | otherwise   = V (VFromW (STpl f) i j)
      conv e = C (SSys$"B"++show rid) (recVariables e++restoreNil)
      widar  = zip wids warities
      widin  = zip wids injs 
      lkup l ls = case lookup l ls of Just v -> v 
      recVariables e = concatMap (\v->
                                  case v of VFromW f i j | not (lkup i widin)->
                                                [V$VFromW (STpl f) i (lkup i widar+j)]
                                            _ ->
                                                []) $ variables e      
      restoreNil = concat [ map V vs | WRule i _ (SNrm f []) vs <- swdecls ] 
      wdecls' = map (\(WRule i vs f vs') ->
                         if f == idSymb then 
                             WRule i vs f vs' -- Special Treatment for ID
                         else 
                             WRule i
                               (lvars f i (length vs*2))
                               (STpl f)
                               vs') swdecls
          where lvars f i len = map (\k -> VFromW (STpl f) i k) [0..len-1]


-- basicComplementation f rules substs limit = 
--     fix [(ARule i f' ps es ws, [0..length es-1]) | ARule i f' ps es ws <- rules, f==f'] 
--         []
--         (maximumRID rules+1)
--         [(ARule i f ps es ws, jdxs) | ARule i f ps es ws <- rules, f==idSymb, jdxs <- [[],[0]]]
--     where fix [] rs cnt proc = rs
--           fix ((r,idxs):reqs) rs cnt proc 
--               | (r,idxs) `elem` proc = fix reqs rs cnt proc
--               | otherwise = 
--                   case basicComplementationStep r idxs rules substs limit cnt proc of
--                     ( r', cnt', reqs') ->
--                         let reqs'' = [ (ARule i f ps vs ws, jdxs) | 
--                                             (f',jdxs) <- reqs', 
--                                             ARule i f ps vs ws <- rules, f==f' ] 
--                         in fix (reqs''++reqs) (r':rs) cnt' ((r,idxs):proc)
                    
                

-- basicComplementationStep (ARule rid f ps es wdecls) idxs rules substs limit cnt proced =
--     (ARule (cnt+1) (STpl f idxs) ps (map chVar (onlyIth es)++es') wdecls', cnt+1, require)
--     where
--       undef = C (SSys "_|_") []
--       swdecls = sortBy (\(WRule i _ _ _) (WRule j _ _ _) -> compare i j) wdecls
--       wids     = [ i | WRule i _ _ _ <- swdecls ]
--       warities = [ length ps | WRule i ps _ _ <- swdecls ]
--       wsymbols = [ f | WRule i _ f _ <- swdecls ]
--       avars = nub $ concat [ activeVars (es !! i) | i <- idxs ]
--       jidxss = map (\wid ->sort$concatMap (\v->
--                                            case v of VFromW _ i j | i == wid -> [j]
--                                                      _ -> []) avars) wids
--       injs  = zipWith (\g jdxs-> detectInjectiveI g jdxs rules substs limit) wsymbols jidxss
--       widjs  = zip wids jidxss
--       widar  = zip wids warities
--       widin  = zip wids injs 
--       lkup l ls = case lookup l ls of Just v -> v 
--       onlyIth es = zipWith(\i e-> if i `elem` idxs then e else undef ) [0..] es
--       es' = zipWith (\i e->if i `elem` idxs then convE e else undef ) [0..] es
--       convE e = C (SSys ("B"++show rid++show idxs)) (recVariables e++infoG e++restoreNil)
--       restoreNil = concat$zipWith (\idxs (WRule _ _ _ vs) -> if null idxs then map V vs else []) jidxss swdecls
--       recVariables e = concatMap (\v->
--                                   case v of VFromW f i j | not (lkup i widin)->
--                                                 [V$VFromW (STpl f (lkup i widjs)) i (lkup i widar+j)]
--                                             _ ->
--                                                 []) $ activeVars e
--       infoG (C _ ts) = concatMap infoG ts
--       infoG (F _ ts) = concatMap infoG ts
--       infoG (V _)    = []
--       chVar (C c ts)              = C c (map chVar ts)
--       chVar (F f ts)              = F f (map chVar ts)
--       chVar (V (VFromW f i j)) | f == idSymb = V (VFromW f i j) -- Special Treatment for ID
--                                | otherwise   = V (VFromW (STpl f (lkup i widjs)) i j)
--       require = nub $ filter (\(f,i) -> f /= idSymb) $zip wsymbols jidxss 
--       wdecls' = zipWith (\(WRule i vs f vs') jdxs->
--                          if f == idSymb then 
--                              WRule i vs f vs' -- Special Treatment for ID
--                          else 
--                              WRule i
--                                (lvars f i jdxs (length vs*2))
--                                (STpl f jdxs)
--                                vs') swdecls jidxss
--           where lvars f i jdxs len = map (\k -> VFromW (STpl f jdxs) i k) [0..len-1]

{-
tplName  f = "#"++f
cmplName f = "!"++f
cmplVar  v = "!"++v

basicComplementation :: Rule a -> [Rule a] -> [Rule a]
basicComplementation i rules = 
    let r = rules !! i 
        Rule (f, PTpl ps, ETpl es, wdecls) = rules !! i
    in tpl [0..(length ps-1)] r [] []
    where ixsToName ixs = "_" ++ foldr (\a r->show a++"_"++r) ixs
          tpl ixs (Rule (f,PTpl ps, ETpl es, wdecls)) defs proced | (f,ps) `elem` proced = (Nothing,(f,ps):proced,[])
                  (Just$ Rule (tplName f++ixsToName ixs, PTpl ps, ETpl (es ++ es'), wdecls'),
                           (zipWith (\WDecl (p,EFun f es) ij -> (f,ij)) wdecls ijs))
                       where collectvars = foldr union [] [ activeVars (es !! i ) | i <- ixs ]
                             varsw = map (\WDecl (p,e) -> vars p) wdecls
                             vjs   = map (\vars -> vars `intersect` collectvars) varsw
                             ijs   = zipWith (\vars ovars -> map (\v -> fromJust$elemIndex v ovars) vars) vjs varsw
                             es'   = zipWith (\e i->if i `elem` ixs then
                                                   ECon ("B"++(length proced+1))
                                                             ((concatMap vars ps \\ collectvars) \\ (concatMap (\WDecl (p,e) -> vars e))
                                                                            `union` (map cmplVar [ activeVars e ]))
                                                        else
                                                            ECon "_|_" []) es [0..(length es-1)]
                             wdecls' = zipWith (\(WDecl (p,EFun f es)) ij -> 
                                                    WDecl (conv p, EFun (tplName f++ixsToName ixs) es)) wdecls ijs
                             conv (PTpl ps) = PTpl (ps++map pv ps)
                             conv (PVar  v) = PTpl ([v,pv v])
                             pv (PVar v) = PVar (cmplVar v)

-}