{
module Lexer where
import Control.Monad.Fix
}

%wrapper "monad"

$alpha_or_number = [a-zA-Z0-9_]
$big_alpha       = [A-Z]
$small_alpha     = [a-z]


@decimal = [1-9] [0-9]*
@ident   = _? $small_alpha $alpha_or_number*
@cident  = _? $big_alpha   $alpha_or_number*

tokens :-
<0> $white+ ; -- We skip whitespaces
<0> "="                  { mkTok $ \s -> TkEq }
<0> "("                  { mkTok $ \s -> TkLPar }
<0> ")"                  { mkTok $ \s -> TkRPar }
<0> "{"                  { mkTok $ \s -> TkLBrace }
<0> "}"                  { mkTok $ \s -> TkRBrace }
<0> "where"              { mkTok $ \s -> TkWhere }
<0> "|"                  { mkTok $ \s -> TkPar }
<0> ":"                  { mkTok $ \s -> TkColon }
<0> ";"                  { mkTok $ \s -> TkSemiColon }
<0> "->"                 { mkTok $ \s -> TkArr }
<0> ","                  { mkTok $ \s -> TkComma }
<0> "@"                  { mkTok $ \s -> TkAt }
<0> "["                  { mkTok $ \s -> TkLBracket }
<0> "]"                  { mkTok $ \s -> TkRBracket }
<0> @ident / $white* "(" { mkTok $ \s -> TkFIdent s }
<0> @ident               { mkTok $ \s -> TkIdent s }
<0> @cident              { mkTok $ \s -> TkCIdent s }
<0> @decimal             { mkTok $ \s -> TkInt (read s) }
<0> .                    { mkTok $ \s -> TkError s }
{
type PToken = ((Int,Int), Token)

-- pos :: PToken -> AlexPosn
-- pos (p,t) = p

body = snd

data Token 
    = TkEq            -- = 
    | TkLPar          -- (
    | TkRPar          -- )
    | TkLBrace        -- {
    | TkRBrace        -- }
    | TkPar           -- |
    | TkColon         -- :
    | TkSemiColon     -- ;
    | TkArr           -- ->
    | TkComma         -- ,
    | TkAt            -- @
    | TkLBracket      -- [
    | TkRBracket      -- ]
    | TkWhere         -- where
    | TkIdent  String -- variables
    | TkFIdent String -- functions 
    | TkCIdent String -- Constructors 
    | TkInt    Int
    | TkEOF
    | TkError String 
      deriving (Eq,Ord)

intValue (TkInt i) = i
strValue (TkIdent s)  = s
strValue (TkCIdent s) = s
strValue (TkFIdent s) = s
strValue _            = ""

instance Show Token where 
    show TkEq    = "="
    show TkLPar  = "("
    show TkRPar  = ")"
    show TkLBrace = "{"
    show TkRBrace = "}"
    show TkWhere  = "where"
    show TkPar    = "|"
    show TkColon  = ":"
    show TkSemiColon = ";"
    show TkArr    = "->"
    show TkComma = ","
    show TkAt    = "@"
    show TkLBracket = "["
    show TkRBracket = "]" 
    show (TkIdent s)  = s
    show (TkCIdent s) = s
    show (TkFIdent s) = s
    show (TkInt i)    = show i



mkTok f (AlexPn p n c,_,str) len = return $ ((n,c), f (take len str))
alexEOF = return $ (undefined, TkEOF)

scanTokens str 
    = runAlex str $ 
      fix (\loop ls -> do 
             { token@(p, tk) <- alexMonadScan;
               case tk of 
                 TkEOF -> return $ reverse ls 
                 _     -> loop $! (token:ls) }) []
}
