{
module Parser where

import ParseTree
import Lexer 
import Control.Monad.Error

}

%name parseProg Prog

%monad { Either String } { (>>=) } { return }
%tokentype { Lexer.PToken } 

%token '='     { (_,TkEq) }
       '('     { (_,TkLPar) }
       ')'     { (_,TkRPar) }
       '{'     { (_,TkLBrace) }
       '}'     { (_,TkRBrace) }
       '['     { (_,TkLBracket) }
       ']'     { (_,TkRBracket) }
       '|'     { (_,TkPar) }
       ':'     { (_,TkColon) }
       ';'     { (_,TkSemiColon) }
       "->"    { (_,TkArr) }
       "where" { (_,TkWhere) }
       ','     { (_,TkComma) }
       '@'     { (_,TkAt) }
       integer { (_,TkInt _) }
       ident   { (_,TkIdent _) }
       fident  { (_,TkFIdent _) }
       cident  { (_,TkCIdent _) }

%right ':'

%%

Prog :: { [Rule] }
: Rule Prog { $1:$2 }
| Rule ';' Prog { $1:$3 }
|               { [] }

Rule :: { Rule }
: fident '(' Pats ')' '=' TE  "where" WDecls { Rule (strValue $ body $1, PTpl $3, $6, $8) }
| fident '(' Pats ')' '=' TE                 { Rule (strValue $ body $1, PTpl $3, $6, []) }

TE :: { Exp }
: '(' TExps ')' { ETpl $2 }
| Exp          { ETpl [$1] }

WDecls :: { [WDecl] }
: WDecl WDecls { $1:$2 }
| WDecl        { [$1] }

WDecl :: { WDecl }
: '(' PVars ')' '=' fident '(' Args ')' { WDecl ( PTpl $2, EFun (strValue $ body $5) $7) }
|     ident     '=' fident '(' Args ')' { WDecl ( PVar (strValue $ body $1), EFun (strValue $ body $3) $5) }

PVars :: { [Pat] }
: ident ',' PVars { PVar (strValue $ body $1):$3 }
| ident           {[PVar (strValue $ body $1)]}

Pats :: { [Pat] }
: Pat ',' Pats { $1:$3 }
| Pat          { [$1] }

Pat :: { Pat }
: cident '(' Pats ')' { PCon (strValue $ body $1) $3 }
| cident              { PCon (strValue $ body $1) [] } 
| '[' ']'             { PCon "Nil" [] }
| Pat ':' Pat         { PCon "Cons" [$1,$3] }
| '(' Pat ')'         { $2 }
| ident               { PVar (strValue $ body $1) }

Exp :: { Exp }
: cident '(' Exps ')'  { ECon (strValue $ body $1) $3 }
| cident               { ECon (strValue $ body $1) [] }
| '[' ']'              { ECon "Nil" [] }
| Exp ':' Exp          { ECon "Cons" [$1,$3] }
| '(' Exp ')'          { $2 }
| ident                { EVar (strValue $ body $1) }
| fident '(' Args ')'  { EFun (strValue $ body $1) $3 }

Exps :: { [Exp] }
: Exp ',' Exps { $1:$3 }
| Exp          { [$1] }


TExps :: { [Exp] }
: Exp ',' TExps { $1:$3 }
| Exp ',' Exp  { [$1,$3] }

Args :: { [Exp] }
: Arg ',' Args { $1:$3 }
| Arg          { [$1] }

Arg :: { Exp }
: ident  {EVar (strValue $ body $1) }

{
happyError tokens =
    if length tokens > 0 then 
	throwError $ "ParseError at " ++ (p tokens) ++ "\n"
    else
        throwError $ "ParseError at " ++ "Unknown Position"
    where p (( (n,c),  t ):_) = "Line: " ++ show n ++ ", Column: "++show c
}
